import os
import torch
import optuna
import shutil

import numpy as np
import cftool.ml.param_utils as pu

from typing import *
from cftool.misc import hash_code
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cftool.misc import Saving
from cftool.misc import LoggingMixin
from cftool.ml.hpo import HPOBase
from cftool.ml.utils import pattern_type
from cftool.ml.utils import scoring_fn_type
from cftool.ml.utils import Metrics
from cftool.ml.utils import Estimator
from cfdata.types import np_float_type
from cfdata.tabular import task_type_type
from cfdata.tabular import parse_task_type
from cfdata.tabular import TaskTypes
from optuna.trial import Trial

from .basic import *
from ..misc.toolkit import *
from .ensemble import Ensemble
from ..types import data_type
from ..pipeline import Pipeline
from ..protocol import DataProtocol
from ..misc._api import SAVING_DELIM
from ..misc.toolkit import inject_mlflow_stuffs


class _TunerResult(NamedTuple):
    model: str
    repeat_result: RepeatResult

    @property
    def pipelines(self) -> List[Pipeline]:
        pipelines = self.repeat_result.pipelines
        if pipelines is None:
            raise ValueError("`pipelines` are not yet generated")
        return pipelines[self.model]

    @property
    def patterns(self) -> List[ModelPattern]:
        patterns = self.repeat_result.patterns
        if patterns is None:
            raise ValueError("`patterns` are not yet generated")
        return patterns[self.model]

    @property
    def weighted_metrics(self) -> Dict[str, np.ndarray]:
        weighted_metrics = {}
        final_results_dict = self.repeat_result.final_results
        if final_results_dict is None:
            raise ValueError("training process is corrupted")
        final_results = final_results_dict[self.model]
        for key in sorted(final_results[0].weighted_scores.keys()):
            local_metrics = [result.weighted_metrics[key] for result in final_results]
            weighted_metrics[key] = np.array(local_metrics, np_float_type)
        return weighted_metrics


class _Tuner(LoggingMixin):
    data_file = "__data__.pt"
    config_name = "__config__"

    def __init__(
        self,
        x: data_type = None,
        y: data_type = None,
        x_cv: data_type = None,
        y_cv: data_type = None,
        task_type: task_type_type = "",
        **kwargs: Any,
    ):
        # `x` will be None if `load` is called
        if x is None:
            return

        kwargs = shallow_copy_dict(kwargs)
        self.has_column_names = None
        if y is not None:
            y, y_cv = map(to_2d, [y, y_cv])
        elif isinstance(x, str):
            data_config = kwargs.get("data_config", {})
            data_config["task_type"] = task_type
            read_config = kwargs.get("read_config", {})
            delim = read_config.get("delim", kwargs.get("delim"))
            if delim is not None:
                read_config["delim"] = delim
            else:
                print(
                    f"{LoggingMixin.warning_prefix}delimiter of the given "
                    "file dataset is not provided, this may cause incorrect parsing"
                )
            if y is not None:
                read_config["y"] = y
            data_protocol = kwargs.get("data_protocol", "tabular")
            tr_data = DataProtocol.make(data_protocol, **data_config)
            tr_data.read(x, **read_config)
            self.has_column_names = tr_data._has_column_names
            y = tr_data.processed.y
            if x_cv is not None:
                if y_cv is None:
                    y_cv = tr_data.transform(x_cv).y
                else:
                    y_cv = tr_data.transform_labels(y_cv)
        else:
            raise ValueError("`x` should be a file when `y` is not provided")

        self.task_type = task_type
        self.x, self.x_cv = x, x_cv
        self.y, self.y_cv = y, y_cv
        self.base_params = shallow_copy_dict(kwargs)

    def make_estimators(
        self,
        metrics: Optional[Union[str, List[str]]],
        repeat_result: Optional[RepeatResult],
    ) -> List[Estimator]:
        if isinstance(metrics, str):
            metrics = [metrics]
        elif metrics is None:
            if self.task_type is None:
                raise ValueError("either `task_type` or `metrics` should be provided")
            if repeat_result is not None:
                if repeat_result.pipelines is None:
                    raise ValueError("pipelines are not provided in `repeat_result`")
                pipeline = list(repeat_result.pipelines.values())[0][0]
                metrics = []
                for metric in sorted(pipeline.trainer.metrics):
                    if metric in Metrics.sign_dict:
                        metrics.append(metric)
                if not metrics:
                    metrics = None
            if metrics is None:
                if parse_task_type(self.task_type) is TaskTypes.CLASSIFICATION:
                    metrics = ["acc", "auc"]
                else:
                    metrics = ["mae", "mse"]
        return list(map(Estimator, metrics))

    def make_data(self) -> Tuple[data_type, data_type, data_type, data_type]:
        x: data_type
        y: data_type
        if isinstance(self.x, str):
            y = y_cv = None
            x, x_cv = self.x, self.x_cv
        else:
            assert isinstance(self.x, (list, np.ndarray))
            assert isinstance(self.y, (list, np.ndarray))
            x, y = self.x.copy(), self.y.copy()
            if self.x_cv is None:
                x_cv = None
            else:
                assert isinstance(self.x_cv, (list, np.ndarray))
                x_cv = self.x_cv.copy()
            if self.y_cv is None:
                y_cv = None
            else:
                assert isinstance(self.y_cv, (list, np.ndarray))
                y_cv = self.y_cv.copy()
        return x, y, x_cv, y_cv

    def train(
        self,
        model: str,
        params: Dict[str, Any],
        num_repeat: int,
        num_parallel: int,
        temp_folder: str,
        *,
        compress: bool = True,
        cuda: Optional[str] = None,
        sequential: Optional[bool] = None,
    ) -> _TunerResult:
        params = update_dict(params, shallow_copy_dict(self.base_params))
        params["verbose_level"] = 0
        params["use_tqdm"] = False
        params["cuda"] = cuda
        x, y, x_cv, y_cv = self.make_data()
        repeat_result = repeat_with(
            x,
            y,
            x_cv,
            y_cv,
            num_repeat=num_repeat,
            num_jobs=num_parallel,
            models=model,
            compress=compress,
            temp_folder=temp_folder,
            predict_config={"contains_labels": True},
            sequential=sequential,
            **params,
        )
        return _TunerResult(model, repeat_result)

    def save(self, export_folder: str) -> "_Tuner":
        Saving.prepare_folder(self, export_folder)

        data_path = os.path.join(export_folder, self.data_file)
        data = {"x": self.x, "y": self.y, "x_cv": self.x_cv, "y_cv": self.y_cv}
        torch.save(data, data_path)

        config = {"task_type": self.task_type, "base_params": self.base_params}
        Saving.save_dict(config, self.config_name, export_folder)

        return self

    @classmethod
    def load(cls, export_folder: str) -> "_Tuner":
        instance = cls()
        data_path = os.path.join(export_folder, cls.data_file)
        data = torch.load(data_path)
        instance.x, instance.y = data["x"], data["y"]
        instance.x_cv, instance.y_cv = data["x_cv"], data["y_cv"]
        config = Saving.load_dict(cls.config_name, export_folder)
        instance.task_type = config["task_type"]
        instance.base_params = config["base_params"]
        return instance


class HPOResult(NamedTuple):
    hpo: HPOBase
    extra_config: Dict[str, Any]

    @property
    def best_param(self) -> Dict[str, Any]:
        param = shallow_copy_dict(self.hpo.best_param)
        return update_dict(param, shallow_copy_dict(self.extra_config))


def _init_extra_config(
    metrics: Optional[Union[str, List[str]]] = None,
    score_weights: Optional[Dict[str, float]] = None,
    extra_config: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    new = extra_config or {}
    new.setdefault("use_timing_context", False)
    if metrics is not None:
        metric_config = new.setdefault("metric_config", {})
        metric_config["types"] = metrics
        if score_weights is not None:
            metric_config["weights"] = score_weights
    return new


default_scoring = "default"


def tune_with(
    x: data_type,
    y: data_type = None,
    x_cv: data_type = None,
    y_cv: data_type = None,
    *,
    model: str = "fcnn",
    hpo_method: str = "bo",
    task_type: task_type_type = "",
    params: Optional[pu.params_type] = None,
    metrics: Optional[Union[str, List[str]]] = None,
    num_jobs: Optional[int] = None,
    num_repeat: int = 5,
    num_parallel: int = 1,
    num_search: int = 10,
    temp_folder: str = "__tmp__",
    score_weights: Optional[Dict[str, float]] = None,
    estimator_scoring_function: Union[str, scoring_fn_type] = default_scoring,
    search_config: Optional[Dict[str, Any]] = None,
    extra_config: Optional[Dict[str, Any]] = None,
    verbose_level: int = 2,
) -> HPOResult:

    if os.path.isdir(temp_folder):
        print(
            f"{LoggingMixin.warning_prefix}'{temp_folder}' already exists, "
            "it will be overwritten"
        )
        shutil.rmtree(temp_folder)

    extra_config = _init_extra_config(metrics, score_weights, extra_config)
    tuner = _Tuner(x, y, x_cv, y_cv, task_type, **extra_config)
    x, y, x_cv, y_cv = tuner.x, tuner.y, tuner.x_cv, tuner.y_cv

    def _creator(_: Any, __: Any, params_: Dict[str, Any]) -> pattern_type:
        num_jobs_ = num_parallel if hpo.is_sequential else 1
        temp_folder_ = os.path.join(temp_folder, hash_code(str(params_)))
        result = tuner.train(model, params_, num_repeat, num_jobs_, temp_folder_)
        if num_repeat <= 1:
            return result.patterns[0]
        return Ensemble.stacking(result.patterns)

    if params is None:
        default_init_param = pu.Any(pu.Choice(values=[None, "truncated_normal"]))
        params = {
            "optimizer": pu.String(pu.Choice(values=["sgd", "rmsprop", "adam"])),
            "optimizer_config": {"lr": pu.Float(pu.Exponential(1e-5, 0.1))},
            "model_config": {
                "default_encoding_configs": {"init_method": default_init_param},
            },
        }

    hpo = HPOBase.make(
        hpo_method,
        _creator,
        params,
        verbose_level=verbose_level,
    )
    if hpo.is_sequential:
        if num_jobs is None:
            num_jobs = 0
        if num_jobs > 1:
            print(
                f"{LoggingMixin.warning_prefix}`num_jobs` is set but hpo is "
                "sequential, please use `num_parallel` instead"
            )
        num_jobs = 0
    if search_config is None:
        search_config = {}
    update_dict(
        {
            "num_retry": 1,
            "num_search": num_search,
            "score_weights": score_weights,
            "estimator_scoring_function": estimator_scoring_function,
        },
        search_config,
    )
    if num_jobs is not None:
        search_config["num_jobs"] = num_jobs
    search_config.setdefault(
        "parallel_logging_folder",
        os.path.join(temp_folder, "__hpo_parallel__"),
    )
    estimators = tuner.make_estimators(metrics, None)
    hpo.search(x, y, estimators, x_cv, y_cv, **search_config)
    return HPOResult(hpo, extra_config)


class OptunaParam(NamedTuple):
    name: str
    values: Any
    dtype: str  # [int | float | categorical]
    config: Optional[Dict[str, Any]] = None

    def pop(self, trial: Trial) -> Any:
        method = getattr(trial, f"suggest_{self.dtype}")
        if self.dtype == "categorical":
            return method(self.name, self.values)
        low, high = self.values
        return method(self.name, low, high, **(self.config or {}))


optuna_params_type = Dict[str, Union[OptunaParam, Dict[str, Any], str]]


class OptunaParamConverter:
    prefix = "[^optuna^]"

    def get_usage(self, k: str) -> Tuple[Optional[str], Optional[str]]:
        if not k.startswith(self.prefix):
            return None, None
        usage_k = k[len(self.prefix) :]
        if not usage_k.startswith("[") or not usage_k.endswith("]"):
            msg = f"special keys must end with '[]' to indicate its usage"
            raise ValueError(msg)
        usage_k = usage_k[1:-1]
        if not usage_k.startswith("[") or "]" not in usage_k:
            return usage_k, None
        user_prefix_end = usage_k.index("]")
        user_prefix = usage_k[1:user_prefix_end]
        usage_k = usage_k[user_prefix_end + 1 :]
        return usage_k, user_prefix

    def convert(self, optuna_params: optuna_params_type) -> optuna_params_type:
        def _inner(d: optuna_params_type, current: dict) -> None:
            for k, v in d.items():
                usage, user_prefix = self.get_usage(k)
                if usage is not None:
                    attr = getattr(self, f"_convert_{usage}", None)
                    if attr is None:
                        raise NotImplementedError(f"unrecognized usage '{usage}' found")
                    current[k] = attr(v)
                    continue
                if isinstance(v, dict):
                    _inner(v, current.setdefault(k, {}))
                    continue
                if isinstance(v, OptunaParam):
                    current[k] = v
                    continue

        new: optuna_params_type = {}
        _inner(optuna_params, new)
        return new

    @staticmethod
    def _convert_ema_decay(value: Any) -> optuna_params_type:
        prefix = value
        use_ema_key = f"{prefix}_use_ema"
        ema_decay_key = f"{prefix}_ema_decay"
        return {
            "use_ema": OptunaParam(use_ema_key, [True, False], "categorical"),
            "ema_decay": OptunaParam(ema_decay_key, [0.0, 1.0], "float"),
        }

    @staticmethod
    def _convert_clip_norm(value: Any) -> optuna_params_type:
        prefix = value
        use_clip_norm_key = f"{prefix}_use_clip_norm"
        clip_norm_key = f"{prefix}_clip_norm"
        ucn_param = OptunaParam(use_clip_norm_key, [True, False], "categorical")
        return {
            "use_clip_norm": ucn_param,
            "clip_norm": OptunaParam(clip_norm_key, [0.25, 1.25], "float"),
        }

    @staticmethod
    def _convert_dropout(value: Any) -> optuna_params_type:
        prefix = value
        use_dropout_key = f"{prefix}_use_dropout"
        drop_ratio_key = f"{prefix}_drop_ratio"
        return {
            "use_dropout": OptunaParam(use_dropout_key, [True, False], "categorical"),
            "drop_ratio": OptunaParam(drop_ratio_key, [0.1, 0.9], "float"),
        }

    @staticmethod
    def _convert_hidden_units(value: Any) -> optuna_params_type:
        # parse
        config: Dict[str, Any] = {}
        prefix, low, high, num_layers, *args = value.split("_")
        low, high, num_layers = map(int, [low, high, num_layers])
        if args:
            if len(args) == 1:
                if args[0] == "log":
                    config["log"] = True
                else:
                    config["step"] = int(args[0])
            else:
                config["log"] = True
                config["step"] = int(args[0])
        # inject
        num_layers_name = f"{prefix}_num_layers"
        rs: optuna_params_type = {
            "num_layers": OptunaParam(num_layers_name, [1, num_layers], "int")
        }
        for i in range(num_layers):
            key = f"hidden_unit_{i}"
            rs[key] = OptunaParam(f"{prefix}_{key}", [low, high], "int", config)
        return rs

    @staticmethod
    def _convert_dndf_config(value: Any) -> optuna_params_type:
        split = value.split("_")
        if len(split) == 3:
            force = False
            prefix, num_tree, tree_depth = split
        else:
            force = True
            assert split[-1] == "force"
            prefix, num_tree, tree_depth = split[:-1]
        num_tree, tree_depth = map(int, [num_tree, tree_depth])
        num_tree = max(4, num_tree)
        tree_depth = max(2, tree_depth)
        num_tree_key = f"{prefix}_num_tree"
        tree_depth_key = f"{prefix}_tree_depth"
        config: optuna_params_type = {
            "num_tree": OptunaParam(num_tree_key, [4, num_tree], "int", {"log": True}),
            "tree_depth": OptunaParam(tree_depth_key, [2, tree_depth], "int"),
        }
        if not force:
            use_dndf_key = f"{prefix}_use_dndf"
            use_dndf_param = OptunaParam(use_dndf_key, [True, False], "categorical")
            config[use_dndf_key] = use_dndf_param
        return config

    @staticmethod
    def _convert_pruner_config(value: Any) -> optuna_params_type:
        prefix = value
        available_methods = ["auto_prune", "surgery", "simplified"]
        use_pruner_key = f"{prefix}_use_pruner"
        method_key = f"{prefix}_prune_method"
        return {
            "use_pruner": OptunaParam(use_pruner_key, [True, False], "categorical"),
            "prune_method": OptunaParam(method_key, available_methods, "categorical"),
        }

    def pop(
        self,
        usage: str,
        value: Dict[str, OptunaParam],
        trial: Trial,
    ) -> Any:
        attr = getattr(self, f"_parse_{usage}", None)
        if attr is None:
            raise NotImplementedError(f"unrecognized usage '{usage}' found")
        return attr(value, trial)

    def parse(self, usage: str, value: Dict[str, Any]) -> Any:
        attr = getattr(self, f"_parse_{usage}", None)
        if attr is None:
            raise NotImplementedError(f"unrecognized usage '{usage}' found")
        return attr(value, None)

    @staticmethod
    def _parse_ema_decay(d: Dict[str, Any], trial: Trial) -> Any:
        use_ema = d["use_ema"]
        if trial is not None:
            use_ema = use_ema.pop(trial)
        if not use_ema:
            return 0.0
        ema_decay = d["ema_decay"]
        if trial is not None:
            ema_decay = ema_decay.pop(trial)
        return ema_decay

    @staticmethod
    def _parse_clip_norm(d: Dict[str, Any], trial: Trial) -> Any:
        use_clip_norm = d["use_clip_norm"]
        if trial is not None:
            use_clip_norm = use_clip_norm.pop(trial)
        if not use_clip_norm:
            return 0.0
        clip_norm = d["clip_norm"]
        if trial is not None:
            clip_norm = clip_norm.pop(trial)
        return clip_norm

    @staticmethod
    def _parse_dropout(d: Dict[str, Any], trial: Trial) -> Any:
        use_dropout = d["use_dropout"]
        if trial is not None:
            use_dropout = use_dropout.pop(trial)
        if not use_dropout:
            return 0.0
        drop_ratio = d["drop_ratio"]
        if trial is not None:
            drop_ratio = drop_ratio.pop(trial)
        return drop_ratio

    @staticmethod
    def _parse_hidden_units(d: Dict[str, Any], trial: Trial) -> Any:
        hidden_units = []
        num_layers = d["num_layers"]
        if trial is not None:
            max_layers = num_layers.values[1]
            num_layers = num_layers.pop(trial)
            for i in range(num_layers, max_layers):
                d.pop(f"hidden_unit_{i}", None)
        for i in range(num_layers):
            key = f"hidden_unit_{i}"
            hidden_unit = d.pop(key, None)
            if hidden_unit is None:
                raise ValueError(f"'{key}' is not found in `model_config`")
            if trial is not None:
                hidden_unit = hidden_unit.pop(trial)
            hidden_units.append(hidden_unit)
        return hidden_units

    @staticmethod
    def _parse_dndf_config(d: Dict[str, Any], trial: Trial) -> Any:
        use_dndf = d.get("use_dndf", None)
        if use_dndf is None:
            use_dndf = True
        else:
            if trial is not None:
                use_dndf = use_dndf.pop(trial)
        if not use_dndf:
            return
        num_tree, tree_depth = d["num_tree"], d["tree_depth"]
        if trial is not None:
            num_tree = num_tree.pop(trial)
            tree_depth = tree_depth.pop(trial)
        return {"num_tree": num_tree, "tree_depth": tree_depth}

    @staticmethod
    def _parse_pruner_config(d: Dict[str, Any], trial: Trial) -> Any:
        use_pruner = d["use_pruner"]
        if trial is not None:
            use_pruner = use_pruner.pop(trial)
        if not use_pruner:
            return
        method = d["prune_method"]
        if trial is not None:
            method = method.pop(trial)
        return {"method": method}

    @staticmethod
    def _make_prefix(prefix: str) -> str:
        return f"[{prefix}]" if prefix else ""

    # api

    @classmethod
    def merge_user_prefix(cls, k: str, user_prefix: Optional[str]) -> str:
        if user_prefix is None:
            return k
        return f"{user_prefix}_{k}"

    @classmethod
    def make_ema_decay(cls, prefix: str) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}ema_decay]"
        value = prefix
        return {key: value}

    @classmethod
    def make_clip_norm(cls, prefix: str) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}clip_norm]"
        value = prefix
        return {key: value}

    @classmethod
    def make_dropout(cls, prefix: str) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}dropout]"
        value = prefix
        return {key: value}

    @classmethod
    def make_hidden_units(
        cls,
        prefix: str,
        low: int,
        high: int,
        num_layers: int,
        step: int = 1,
        log: bool = True,
    ) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}hidden_units]"
        value = f"{prefix}_{low}_{high}_{num_layers}_{step}"
        if log:
            value = f"{value}_log"
        return {key: value}

    @classmethod
    def make_dndf_config(
        cls,
        prefix: str,
        num_tree: int,
        tree_depth: int,
        force: bool,
    ) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}dndf_config]"
        value = f"{prefix}_{num_tree}_{tree_depth}"
        if force:
            value = f"{value}_force"
        return {key: value}

    @classmethod
    def make_pruner_config(cls, prefix: str) -> Dict[str, str]:
        key = f"{cls.prefix}[{cls._make_prefix(prefix)}pruner_config]"
        value = prefix
        return {key: value}


class OptunaKeyMapping(LoggingMixin):
    tuner_folder = "__tuner__"
    optuna_params_name = "__optuna_params__"

    def __init__(self, tuner: _Tuner, optuna_params: optuna_params_type):
        self.tuner = tuner
        self.delim = SAVING_DELIM
        self.converter = OptunaParamConverter()
        self.params = self.converter.convert(optuna_params)
        self.optuna_key_mapping: Dict[str, str] = {}

        def _inject_mapping(d: optuna_params_type, prefix_list: List[str]) -> None:
            for k, v in d.items():
                new_prefix_list = prefix_list + [k]
                if isinstance(v, OptunaParam):
                    self.optuna_key_mapping[v.name] = self.delim.join(new_prefix_list)
                    continue
                assert isinstance(v, dict)
                _inject_mapping(v, new_prefix_list)

        _inject_mapping(self.params, [])
        self._optuna_params = optuna_params

    def pop(self, trial: Trial) -> Dict[str, Any]:
        optuna_params = shallow_copy_dict(self.params)
        current_params = shallow_copy_dict(self.tuner.base_params)

        def _inject_suggestion(d: optuna_params_type, current: Dict[str, Any]) -> None:
            for k, v in d.items():
                usage, user_prefix = self.converter.get_usage(k)
                if usage is not None:
                    assert isinstance(v, dict)
                    k = OptunaParamConverter.merge_user_prefix(usage, user_prefix)
                    current[k] = self.converter.pop(usage, v, trial)
                    continue
                if isinstance(v, dict):
                    _inject_suggestion(v, current.setdefault(k, {}))
                    continue
                assert isinstance(v, OptunaParam)
                current[k] = v.pop(trial)

        _inject_suggestion(optuna_params, current_params)
        return current_params

    def parse(self, optuna_param_values: Dict[str, Any]) -> Dict[str, Any]:
        params: Dict[str, Any] = {}
        for k, v in optuna_param_values.items():
            key_mapping = self.optuna_key_mapping[k]
            key_path = key_mapping.split(self.delim)
            local_param = params
            for sub_k in key_path[:-1]:
                local_param = local_param.setdefault(sub_k, {})
            local_param[key_path[-1]] = v
        return params

    def convert(self, param_values: Dict[str, Any]) -> Dict[str, Any]:
        converted: Dict[str, Any] = {}

        def _inject_values(d: Dict[str, Any], current: dict) -> None:
            for k, v in d.items():
                usage, user_prefix = self.converter.get_usage(k)
                if usage is not None:
                    current[usage] = self.converter.parse(usage, v)
                    continue
                if isinstance(v, dict):
                    _inject_values(v, current.setdefault(k, {}))
                    continue
                current[k] = v

        _inject_values(param_values, converted)
        return converted

    def save(self, export_folder: str) -> "OptunaKeyMapping":
        Saving.prepare_folder(self, export_folder)
        tuner_folder = os.path.join(export_folder, self.tuner_folder)
        self.tuner.save(tuner_folder)
        Saving.save_dict(self._optuna_params, self.optuna_params_name, export_folder)
        return self

    @classmethod
    def load(cls, export_folder: str) -> "OptunaKeyMapping":
        tuner_folder = os.path.join(export_folder, cls.tuner_folder)
        tuner = _Tuner.load(tuner_folder)
        optuna_params = Saving.load_dict(cls.optuna_params_name, export_folder)
        return cls(tuner, optuna_params)


class OptunaResult(NamedTuple):
    tuner: _Tuner
    study: optuna.study.Study
    optuna_key_mapping: OptunaKeyMapping

    @property
    def best_param(self) -> Dict[str, Any]:
        param = shallow_copy_dict(self.tuner.base_params)
        optuna_param = self.optuna_key_mapping.parse(self.study.best_params)
        update_dict(optuna_param, param)
        return self.optuna_key_mapping.convert(param)


class OptunaPresetParams:
    def __init__(
        self,
        *,
        tune_lr: bool = True,
        tune_optimizer: bool = True,
        tune_scheduler: bool = True,
        tune_ema_decay: bool = True,
        tune_clip_norm: bool = True,
        tune_batch_size: bool = True,
        tune_init_method: bool = True,
        **kwargs: Any,
    ) -> None:
        self.base_params: optuna_params_type = {}
        if tune_lr:
            lr_param = OptunaParam("lr", [1e-5, 0.1], "float", {"log": True})
            optimizer_config = self.base_params.setdefault("optimizer_config", {})
            assert isinstance(optimizer_config, dict)
            optimizer_config["lr"] = lr_param
        if tune_optimizer:
            optimizer_param = OptunaParam(
                "optimizer",
                ["nag", "rmsprop", "adam", "adamw"],
                "categorical",
            )
            self.base_params["optimizer"] = optimizer_param
        if tune_scheduler:
            scheduler_param = OptunaParam(
                "scheduler",
                ["step", "plateau", "warmup"],
                "categorical",
            )
            self.base_params["scheduler"] = scheduler_param
        if tune_ema_decay:
            model_config = self.base_params.setdefault("model_config", {})
            assert isinstance(model_config, dict)
            model_config.update(OptunaParamConverter.make_ema_decay(""))
        if tune_clip_norm:
            trainer_config = self.base_params.setdefault("trainer_config", {})
            assert isinstance(trainer_config, dict)
            trainer_config.update(OptunaParamConverter.make_clip_norm(""))
        if tune_batch_size:
            choices = [16, 32, 64, 128, 256, 512, 1024]
            bs_param = OptunaParam("batch_size", choices, "categorical")
            self.base_params["batch_size"] = bs_param
        if tune_init_method:
            default_encoding_init_param = OptunaParam(
                "default_encoding_init_method",
                [None, "truncated_normal"],
                "categorical",
            )
            model_config = self.base_params.setdefault("model_config", {})
            assert isinstance(model_config, dict)
            de_cfg = model_config.setdefault("default_encoding_configs", {})
            de_cfg["init_method"] = default_encoding_init_param
        self.kwargs = shallow_copy_dict(kwargs)

    def get(self, model: str) -> optuna_params_type:
        attr = getattr(self, f"_{model}_preset", None)
        if attr is None:
            raise NotImplementedError(f"preset params for '{model}' is not defined")
        preset = attr()
        if not preset:
            raise ValueError("current preset params is empty")
        return preset

    def _nnb_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)

    def _ndt_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)

    def _linear_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)

    @staticmethod
    def _get_head_config(params: Dict[str, Any], pipe: str) -> Dict[str, Any]:
        model_config = params.setdefault("model_config", {})
        pipe_configs = model_config.setdefault("pipe_configs", {})
        return pipe_configs.setdefault(pipe, {}).setdefault("head", {})

    def _fcnn_preset(self) -> optuna_params_type:
        params = shallow_copy_dict(self.base_params)
        if self.kwargs.get("tune_hidden_units", True):
            hu_param = OptunaParamConverter.make_hidden_units("", 8, 2048, 3)
            head_config = self._get_head_config(params, "fcnn")
            head_config.update(hu_param)
        mapping_config: optuna_params_type = {}
        if self.kwargs.get("tune_batch_norm", True):
            bn_param = OptunaParam("mlp_batch_norm", [False, True], "categorical")
            mapping_config["batch_norm"] = bn_param
        if self.kwargs.get("tune_dropout", True):
            mapping_config.update(OptunaParamConverter.make_dropout(""))
        if self.kwargs.get("tune_pruner", True):
            mapping_config.update(OptunaParamConverter.make_pruner_config(""))
        if mapping_config:
            head_config = self._get_head_config(params, "fcnn")
            head_config["mapping_configs"] = mapping_config
        if self.kwargs.get("tune_embedding_dim", True):
            ed_param = OptunaParam("embedding_dim", [8, "auto"], "categorical")
            model_config = params.setdefault("model_config", {})
            assert isinstance(model_config, dict)
            model_config["default_encoding_configs"]["embedding_dim"] = ed_param
        return params

    def _tree_dnn_preset(self) -> optuna_params_type:
        params = self._fcnn_preset()
        if self.kwargs.get("tune_dndf", True):
            dndf_param = OptunaParamConverter.make_dndf_config("", 32, 4, False)
            head_config = self._get_head_config(params, "dndf")
            head_config.update(dndf_param)
        return params

    def _tree_linear_preset(self) -> optuna_params_type:
        params = shallow_copy_dict(self.base_params)
        if self.kwargs.get("tune_dndf", True):
            dndf_param = OptunaParamConverter.make_dndf_config("out", 64, 4, True)
            head_config = self._get_head_config(params, "tree_stack")
            head_config.update(dndf_param)
        return params

    def _tree_stack_preset(self) -> optuna_params_type:
        params = shallow_copy_dict(self.base_params)
        if self.kwargs.get("tune_num_blocks", True):
            head_config = self._get_head_config(params, "tree_stack")
            head_config["num_blocks"] = OptunaParam("num_blocks", [1, 3], "int")
        if self.kwargs.get("tune_inner_dndf", True):
            head_config = self._get_head_config(params, "tree_stack")
            inner_param = OptunaParamConverter.make_dndf_config("", 64, 4, True)
            head_config.update(inner_param)
        if self.kwargs.get("tune_dndf", True):
            head_config = self._get_head_config(params, "tree_stack")
            out_param = OptunaParamConverter.make_dndf_config("out", 32, 4, True)
            head_config.update(out_param)
        return params

    # TODO : optimize these three preset

    def _ddr_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)

    def _ddr_q_preset(self) -> optuna_params_type:
        return self._ddr_preset()

    def _ddr_cdf_preset(self) -> optuna_params_type:
        return self._ddr_preset()

    def _rnn_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)

    def _transformer_preset(self) -> optuna_params_type:
        return shallow_copy_dict(self.base_params)


class OptunaArgs(NamedTuple):
    cuda: Optional[str]
    compress: bool
    num_trial: Union[str, int]
    task_config: Union[str, Dict[str, Any]]
    key_mapping: Union[str, OptunaKeyMapping]


def optuna_core(args: OptunaArgs) -> optuna.study.Study:
    cuda = args.cuda
    if cuda == "None":
        cuda = None
    compress = args.compress

    key_mapping_arg = args.key_mapping
    if isinstance(key_mapping_arg, str):
        key_mapping = OptunaKeyMapping.load(key_mapping_arg)
    else:
        assert isinstance(key_mapping_arg, OptunaKeyMapping)
        key_mapping = key_mapping_arg
    tuner = key_mapping.tuner

    config = args.task_config
    if isinstance(config, dict):
        config_dict = config
    else:
        config_dict = Saving.load_dict("config", config)
    model = config_dict["model"]
    metrics = config_dict["metrics"]
    timeout = config_dict["timeout"]
    num_jobs = config_dict["num_jobs"]
    num_repeat = config_dict["num_repeat"]
    num_parallel = config_dict["num_parallel"]
    estimator_scoring_function = config_dict["estimator_scoring_function"]
    study_config = config_dict["study_config"]
    temp_folder = config_dict["temp_folder"]

    def objective(trial: Trial) -> float:
        temp_folder_ = os.path.join(temp_folder, str(trial.number))
        current_params = key_mapping.pop(trial)
        if metrics is not None:
            current_params["metrics"] = metrics
        inject_mlflow_stuffs(
            model,
            config=current_params,
            run_name_prefix=f"trial_{trial.number}",
        )
        args_ = model, current_params, num_repeat, num_parallel, temp_folder_
        sequential = None if num_jobs <= 1 else True
        result = tuner.train(
            *args_,
            sequential=sequential,
            compress=compress,
            cuda=cuda,
        )
        final_scores = []
        weighted_metrics = result.weighted_metrics
        estimators = tuner.make_estimators(metrics, result.repeat_result)
        for estimator in estimators:
            scoring_fn = estimator.scoring_fn(estimator_scoring_function)
            final_scores.append(scoring_fn(weighted_metrics[estimator.type]))
        return sum(final_scores) / len(final_scores)

    study = optuna.create_study(**study_config)
    study.optimize(objective, int(args.num_trial), timeout, 1)
    return study


def optuna_tune(
    x: data_type = None,
    y: data_type = None,
    x_cv: data_type = None,
    y_cv: data_type = None,
    *,
    model: str = "fcnn",
    task_type: task_type_type = "",
    tuner: Optional[_Tuner] = None,
    params: Optional[optuna_params_type] = None,
    study_config: Optional[Dict[str, Any]] = None,
    metrics: Optional[Union[str, List[str]]] = None,
    num_jobs: int = 1,
    num_trial: int = 50,
    num_repeat: int = 5,
    num_parallel: int = 1,
    compress: bool = True,
    temp_folder: str = "__tmp__",
    estimator_scoring_function: Union[str, scoring_fn_type] = default_scoring,
    timeout: Optional[float] = None,
    score_weights: Optional[Dict[str, float]] = None,
    extra_config: Optional[Dict[str, Any]] = None,
    cuda: Optional[Union[int, str]] = None,
) -> OptunaResult:
    if params is None:
        params = OptunaPresetParams().get(model)

    if tuner is None:
        if x is None:
            raise ValueError("either `x` or `tuner` should be provided")
        extra_config = _init_extra_config(metrics, score_weights, extra_config)
        tuner = _Tuner(x, y, x_cv, y_cv, task_type, **extra_config)
    key_mapping = OptunaKeyMapping(tuner, params)

    if num_jobs <= 1:
        meta_folder = None
    else:
        if cuda is not None:
            print(
                f"{LoggingMixin.warning_prefix}`cuda` is set to {cuda} "
                "but will take no effect because `num_jobs` is set "
                f"to be greater than 1 ({num_jobs})"
            )
        meta_folder = os.path.join(temp_folder, "__meta__")

    if study_config is None:
        study_config = {}
    if num_jobs <= 1:
        storage = None
    else:
        assert isinstance(meta_folder, str)
        os.makedirs(meta_folder, exist_ok=True)
        storage = os.path.join(meta_folder, "shared.db")
        storage = f"sqlite:///{storage}"
    study_config["storage"] = storage
    study_config["direction"] = "maximize"
    study_config["load_if_exists"] = True
    study_config.setdefault("study_name", f"{model}_optuna")
    if num_jobs > 1:
        optuna.create_study(**study_config)

    task_config = {
        "model": model,
        "metrics": metrics,
        "timeout": timeout,
        "num_jobs": num_jobs,
        "num_repeat": num_repeat,
        "num_parallel": num_parallel,
        "estimator_scoring_function": estimator_scoring_function,
        "study_config": study_config,
        "temp_folder": temp_folder,
    }

    if num_jobs <= 1:
        if isinstance(cuda, int):
            cuda = str(cuda)
        args = OptunaArgs(cuda, compress, num_trial, task_config, key_mapping)
        study = optuna_core(args)
    else:
        assert isinstance(meta_folder, str)

        key_mapping_folder = os.path.join(meta_folder, "__key_mapping__")
        key_mapping.save(key_mapping_folder)

        task_config_folder = os.path.join(meta_folder, "__task_config__")
        os.makedirs(task_config_folder, exist_ok=True)
        Saving.save_dict(task_config, "config", task_config_folder)

        experiment = Experiment(num_jobs=num_jobs)
        num_trials = [num_trial // num_jobs] * num_jobs
        num_trials[-1] = num_trial - sum(num_trials[:-1])
        for n in num_trials:
            experiment.add_task(
                execute="optuna",
                root_workplace=temp_folder,
                num_trial=n,
                compress=compress,
                task_config_folder=task_config_folder,
                key_mapping_folder=key_mapping_folder,
            )
        experiment.run_tasks()
        study = optuna.create_study(**study_config)

    return OptunaResult(tuner, study, key_mapping)


__all__ = [
    "tune_with",
    "optuna_core",
    "optuna_tune",
    "OptunaParam",
    "OptunaParamConverter",
    "OptunaKeyMapping",
    "OptunaPresetParams",
]



import torch

import torch.nn as nn

from typing import *

from ..base import ExtractorBase
from ...transform.core import Dimensions
from ....misc.toolkit import Activations
from ....modules.blocks import Linear
from ....modules.blocks import Dropout
from ....modules.blocks import Attention


class TransformerLayer(nn.Module):
    def __init__(
        self,
        input_dim: int,
        num_heads: int,
        *,
        dropout: float = 0.1,
        latent_dim: int = 2048,
        activation: str = "ReLU",
        attention_config: Optional[Dict[str, Any]] = None,
        activation_config: Optional[Dict[str, Any]] = None,
        to_latent_config: Optional[Dict[str, Any]] = None,
        from_latent_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        if attention_config is None:
            attention_config = {}
        attention_config["is_self_attention"] = True
        attention_config.setdefault("dropout", dropout)
        self.self_attn = Attention(input_dim, num_heads, **attention_config)
        if to_latent_config is None:
            to_latent_config = {}
        self.to_latent = Linear(input_dim, latent_dim, **to_latent_config)
        self.dropout = Dropout(dropout)
        if from_latent_config is None:
            from_latent_config = {}
        self.from_latent = Linear(latent_dim, input_dim, **from_latent_config)

        self.norm1 = nn.LayerNorm(input_dim)
        self.norm2 = nn.LayerNorm(input_dim)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.activation = Activations.make(activation, activation_config)

    def forward(self, net: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        new = self.self_attn(net, net, net, mask=mask).output
        net = net + self.dropout1(new)
        net = self.norm1(net)
        new = self.from_latent(self.dropout(self.activation(self.to_latent(net))))
        net = net + self.dropout2(new)
        net = self.norm2(net)
        return net


@ExtractorBase.register("transformer")
class Transformer(ExtractorBase):
    def __init__(
        self,
        in_flat_dim: int,
        dimensions: Dimensions,
        num_heads: int,
        num_layers: int,
        latent_dim: int,
        norm: Optional[Callable],
        input_linear_config: Dict[str, Any],
        transformer_layer_config: Dict[str, Any],
    ):
        super().__init__(in_flat_dim, dimensions)
        # latent projection
        self.input_linear = Linear(self.in_dim, latent_dim, **input_linear_config)
        self.latent_dim = latent_dim
        # transformer blocks
        self.norm = norm
        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    latent_dim,
                    num_heads,
                    **transformer_layer_config,
                )
                for _ in range(num_layers)
            ]
        )

    @property
    def flatten_ts(self) -> bool:
        return False

    @property
    def out_dim(self) -> int:
        return self.latent_dim

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        if self.input_linear is not None:
            net = self.input_linear(net)
        for layer in self.layers:
            net = layer(net, mask=None)
        if self.norm is not None:
            net = self.norm(net)
        return net[..., -1, :]


__all__ = ["Transformer"]



import torch

import numpy as np

from typing import *
from abc import ABCMeta
from torch import Tensor
from torch.nn import Module
from torch.nn import ModuleDict
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cftool.misc import register_core
from cftool.misc import context_error_handler
from cfdata.tabular import TaskTypes
from cfdata.tabular import ColumnTypes

try:
    amp: Optional[Any] = torch.cuda.amp
except:
    amp = None

from ..modules import *
from ..types import tensor_dict_type
from ..losses import LossBase
from ..configs import Configs
from ..configs import Environment
from ..protocol import TrainerState
from ..protocol import ModelProtocol
from ..protocol import DataLoaderProtocol
from ..misc.toolkit import to_torch
from ..modules.heads import HeadBase
from ..modules.heads import HeadConfigs
from ..modules.blocks import DNDF
from ..modules.transform import transform_config_mapping
from ..modules.transform import Transform
from ..modules.transform import Dimensions
from ..modules.transform import SplitFeatures
from ..modules.extractors import ExtractorBase
from ..modules.aggregators import AggregatorBase

model_dict: Dict[str, Type["ModelBase"]] = {}


class PipeConfig(NamedTuple):
    transform: str
    extractor: str
    reuse_extractor: bool
    head: str
    extractor_config: str
    head_config: str
    extractor_meta_scope: Optional[str]
    head_meta_scope: Optional[str]

    @property
    def use_extractor_meta(self) -> bool:
        return self.extractor_meta_scope is not None

    @property
    def use_head_meta(self) -> bool:
        return self.head_meta_scope is not None

    @property
    def extractor_scope(self) -> str:
        if self.extractor_meta_scope is None:
            return self.extractor
        return self.extractor_meta_scope

    @property
    def head_scope(self) -> str:
        if self.head_meta_scope is None:
            return self.head
        return self.head_meta_scope

    @property
    def extractor_config_key(self) -> str:
        if self.extractor_meta_scope is None:
            prefix = self.extractor
        else:
            prefix = self.extractor_meta_scope
        return f"{self.transform}_{prefix}_{self.extractor_config}"

    @property
    def extractor_unique_key(self) -> str:
        prefix = f"{self.extractor_meta_scope}_{self.extractor}"
        return f"{self.transform}_{prefix}_{self.extractor_config}"

    @property
    def head_config_key(self) -> str:
        if self.head_meta_scope is None:
            prefix = self.head
        else:
            prefix = self.head_meta_scope
        return f"{prefix}_{self.head_config}"


class ModelBase(ModelProtocol, metaclass=ABCMeta):
    registered_pipes: Optional[Dict[str, PipeConfig]] = None
    registered_meta_configs: Optional[Dict[str, Dict[str, Any]]] = None

    def __init__(
        self,
        environment: Environment,
        tr_loader: DataLoaderProtocol,
        cv_loader: Optional[DataLoaderProtocol],
        tr_weights: Optional[np.ndarray],
        cv_weights: Optional[np.ndarray],
        loaded_registered_pipes: Optional[Dict[str, PipeConfig]] = None,
    ):
        super().__init__()
        # common
        self.environment = environment
        self.device = environment.device
        self.timing = environment.use_timing_context
        self.tr_loader = tr_loader
        self.cv_loader = cv_loader
        self.data = self.tr_data = tr_loader.data
        self.cv_data = None if cv_loader is None else cv_loader.data
        self.num_train = len(self.tr_data)
        self.num_valid = None if self.cv_data is None else len(self.cv_data)
        self.tr_weights, self.cv_weights = tr_weights, cv_weights
        self._preset_config()
        self._init_config()
        # loss
        self.loss_name = self.config.setdefault("loss", "auto")
        if self.loss_name == "auto":
            if self.tr_data.is_reg:
                self.loss_name = "mae"
            else:
                self.loss_name = "focal"
        loss_config = self.config.setdefault("loss_config", {})
        self.loss = LossBase.make(self.loss_name, loss_config, "none")
        # encoder
        excluded = 0
        numerical_columns_mapping = {}
        categorical_columns_mapping = {}
        categorical_dims = []
        encoding_methods = []
        encoding_configs = []
        true_categorical_columns = []
        if self.tr_data.is_simplify:
            for idx in range(self.tr_data.raw_dim):
                numerical_columns_mapping[idx] = idx
        else:
            ts_indices = self.tr_data.ts_indices
            recognizers = self.tr_data.recognizers
            sorted_indices = [idx for idx in sorted(recognizers) if idx != -1]
            for idx in sorted_indices:
                recognizer = recognizers[idx]
                assert recognizer is not None
                if not recognizer.info.is_valid or idx in ts_indices:
                    excluded += 1
                elif recognizer.info.column_type is ColumnTypes.NUMERICAL:
                    numerical_columns_mapping[idx] = idx - excluded
                else:
                    str_idx = str(idx)
                    categorical_dims.append(recognizer.num_unique_values)
                    encoding_methods.append(
                        self.encoding_methods.setdefault(
                            str_idx, self.default_encoding_method
                        )
                    )
                    encoding_configs.append(
                        self.encoding_configs.setdefault(
                            str_idx, self.default_encoding_configs
                        )
                    )
                    true_idx = idx - excluded
                    true_categorical_columns.append(true_idx)
                    categorical_columns_mapping[idx] = true_idx
        if not true_categorical_columns:
            self.encoder = None
        else:
            loaders = {"tr": self.tr_loader}
            if self.cv_loader is not None:
                loaders["cv"] = self.cv_loader
            encoder_config = self.config.setdefault("encoder_config", {})
            self.encoder = Encoder(
                encoder_config,
                categorical_dims,
                encoding_methods,
                encoding_configs,
                true_categorical_columns,
                loaders,
            )
        # dimensions
        self.dimensions = Dimensions(
            self.encoder,
            numerical_columns_mapping,
            categorical_columns_mapping,
            self.num_history,
        )
        # pipes
        self.pipes: Dict[str, Tuple[str, str, str]] = {}
        self.transforms = ModuleDict()
        self.extractors = ModuleDict()
        self.heads = ModuleDict()
        self.bypassed_pipes: Set[str] = set()
        self._extractor_configs: Dict[str, Dict[str, Any]] = {}
        self._head_configs: Dict[str, Dict[str, Any]] = {}
        self._head_config_ins_dict: Dict[str, HeadConfigs] = {}
        if loaded_registered_pipes is not None:
            self.registered_pipes = loaded_registered_pipes
        if self.registered_pipes is None:
            raise ValueError(f"No `pipe` is registered in {type(self).__name__}")
        with torch.no_grad():
            for key, pipe_config in self.registered_pipes.items():
                local_configs = self.pipe_configs.setdefault(key, {})
                transform_config = local_configs.setdefault("transform", {})
                extractor_config = local_configs.setdefault("extractor", {})
                head_config = local_configs.setdefault("head", {})
                self.add_pipe(
                    key,
                    pipe_config,
                    transform_config,
                    extractor_config,
                    head_config,
                )
        aggregator = self.config["aggregator"]
        aggregator_config = self.config["aggregator_config"]
        self.aggregator = AggregatorBase.make(aggregator, **aggregator_config)
        # caches
        self._transform_cache: Dict[str, Tensor] = {}
        self._extractor_cache: Dict[str, Tensor] = {}

    def __getattr__(self, item: str) -> Any:
        try:
            return super().__getattr__(item)
        except AttributeError:
            value = self.config.get(item)
            if value is not None:
                return value
            value = self.environment.config.get(item)
            if value is not None:
                return value
            msg = f"attribute '{item}' is not defined in {type(self).__name__}"
            raise AttributeError(msg)

    @property
    def config(self) -> Dict[str, Any]:
        return self.environment.model_config

    @property
    def task_type(self) -> TaskTypes:
        return self.tr_data.task_type

    @property
    def labels_key(self) -> str:
        return self.tr_loader.labels_key

    @property
    def num_history(self) -> int:
        num_history = 1
        if self.tr_data.is_ts:
            sampler_config = self.environment.sampler_config
            aggregation_config = sampler_config.get("aggregation_config", {})
            num_history = aggregation_config.get("num_history")
            if num_history is None:
                raise ValueError(
                    "please provide `num_history` in `aggregation_config` "
                    "in `cflearn.make` for time series tasks."
                )
        return num_history

    @property
    def pipe_configs(self) -> Dict[str, Any]:
        return self.config.setdefault("pipe_configs", {})

    def get_pipe_config(self, pipe: str, part: str) -> Dict[str, Any]:
        return self.pipe_configs.setdefault(pipe, {}).setdefault(part, {})

    def add_pipe(
        self,
        key: str,
        pipe_config: PipeConfig,
        transform_config: Dict[str, bool],
        extractor_config: Dict[str, Any],
        head_config: Dict[str, Any],
    ) -> None:
        # meta config
        meta_configs = shallow_copy_dict(self.registered_meta_configs or {})
        meta_config = shallow_copy_dict(meta_configs.get(key, {}))
        meta_transform_config = shallow_copy_dict(meta_config.get("transform", {}))
        meta_extractor_config = shallow_copy_dict(meta_config.get("extractor", {}))
        meta_head_config = shallow_copy_dict(meta_config.get("head", {}))
        # update user increment config
        if meta_config:
            user_increment_config = self.environment.user_increment_config
            user_model_config = user_increment_config.setdefault("model_config", {})
            user_pipe_configs = user_model_config.setdefault("pipe_configs", {})
            user_pipe_config = user_pipe_configs.setdefault(key, {})
            keys = ["transform", "extractor", "head"]
            metas = [meta_transform_config, meta_extractor_config, meta_head_config]
            for k, meta in zip(keys, metas):
                update_dict(meta, user_pipe_config.setdefault(k, {}))
        # transform
        transform_key = pipe_config.transform
        if transform_key in self.transforms:
            transform_exists = True
            transform = self.transforms[transform_key]
        else:
            transform_exists = False
            transform_cfg = Configs.get("transform", transform_key, **transform_config)
            transform_kwargs = transform_cfg.pop()
            update_dict(meta_transform_config, transform_kwargs)
            transform = Transform(self.dimensions, **transform_kwargs)
        # extractor
        extractor_cfg_key = pipe_config.extractor_config_key
        extractor_unique_key = pipe_config.extractor_unique_key
        extractor_exists = extractor_unique_key in self.extractors
        if pipe_config.reuse_extractor and extractor_exists:
            extractor = self.extractors[extractor_unique_key]
        else:
            if extractor_exists:
                new_index = 1
                new_extractor_unique_key = extractor_unique_key
                while new_extractor_unique_key in self.extractors:
                    new_extractor_unique_key = f"{extractor_unique_key}_{new_index}"
                extractor_unique_key = new_extractor_unique_key
            if extractor_cfg_key in self._extractor_configs:
                extractor_kwargs = self._extractor_configs[extractor_cfg_key]
            else:
                extractor_cfg = Configs.get(
                    pipe_config.extractor_scope,
                    pipe_config.extractor_config,
                    **extractor_config,
                )
                extractor_kwargs = extractor_cfg.pop()
                self._extractor_configs[extractor_cfg_key] = extractor_kwargs
            if pipe_config.use_extractor_meta:
                extractor_kwargs = extractor_kwargs[pipe_config.extractor]
            update_dict(meta_extractor_config, extractor_kwargs)
            extractor = ExtractorBase.make(
                pipe_config.extractor,
                transform.out_dim,
                transform.dimensions,
                extractor_kwargs,
            )
        # head
        head_cfg_key = pipe_config.head_config_key
        if head_cfg_key in self._head_configs:
            head_kwargs = self._head_configs[head_cfg_key]
            head_cfg = self._head_config_ins_dict[head_cfg_key]
            head_cfg.in_dim = extractor.out_dim
        else:
            head_cfg = HeadConfigs.get(
                pipe_config.head_scope,
                pipe_config.head_config,
                in_dim=extractor.out_dim,
                tr_data=self.tr_data,
                tr_weights=self.tr_weights,
                dimensions=self.dimensions,
                **head_config,
            )
            head_kwargs = head_cfg.pop()
            self._head_configs[head_cfg_key] = head_kwargs
            self._head_config_ins_dict[head_cfg_key] = head_cfg
        if pipe_config.use_head_meta:
            head_kwargs = head_kwargs[pipe_config.head]
        head_cfg.inject_dimensions(head_kwargs)
        update_dict(meta_head_config, head_kwargs)
        head = HeadBase.make(pipe_config.head, head_kwargs)
        # gather
        self.pipes[key] = transform_key, extractor_unique_key, head_cfg_key
        if not transform_exists:
            self.transforms[transform_key] = transform
        self.extractors[extractor_unique_key] = extractor
        self.heads[key] = head
        # bypass
        if transform.out_dim == 0 or extractor.out_dim == 0:
            self.bypassed_pipes.add(key)

    # Inheritance

    @property
    def input_sample(self) -> tensor_dict_type:
        sample = next(iter(self.tr_loader))
        if self.tr_loader.return_indices:
            assert isinstance(sample, tuple)
            return sample[0]
        assert isinstance(sample, dict)
        return sample

    @property
    def output_probabilities(self) -> bool:
        return False

    def _preset_config(self) -> None:
        pass

    def _init_config(self) -> None:
        default_encoding_method = set()
        if self.registered_pipes is None:
            raise ValueError(f"No `pipe` is registered in {type(self).__name__}")
        for pipe_config in self.registered_pipes.values():
            transform_config = transform_config_mapping[pipe_config.transform]
            if transform_config["one_hot"]:
                default_encoding_method.add("one_hot")
            if transform_config["embedding"]:
                default_encoding_method.add("embedding")
        self.default_encoding_method = self.config.setdefault(
            "default_encoding_method", list(default_encoding_method)
        )
        self.loss_config["input_logits"] = not self.output_probabilities

    def forward(
        self,
        batch: tensor_dict_type,
        batch_idx: Optional[int] = None,
        state: Optional[TrainerState] = None,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        # batch will have `categorical`, `numerical` and `labels` keys
        x_batch = batch["x_batch"]
        split = self._split_features(x_batch, batch_indices, loader_name)
        outputs = self.execute(split)
        return self.aggregator.reduce(outputs, **kwargs)

    def loss_function(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        batch_indices: Optional[torch.Tensor],
        forward_results: tensor_dict_type,
        state: Optional[TrainerState],
    ) -> tensor_dict_type:
        # requires returning `loss` key
        labels = batch[self.labels_key]
        # `sample_weights` could be accessed through:
        # 1) `self.tr_weights[batch_indices]` (for training)
        # 2) `self.cv_weights[batch_indices]` (for validation)
        losses = self.loss(forward_results, labels)
        if isinstance(losses, dict):
            return {k: v.mean() for k, v in losses.items()}
        return {"loss": losses.mean()}

    # API

    @property
    def configurations(self) -> Dict[str, Any]:
        return self.environment.config

    def _split_features(
        self,
        x_batch: Tensor,
        batch_indices: Optional[np.ndarray],
        loader_name: Optional[str],
    ) -> SplitFeatures:
        return self.dimensions.split_features(
            x_batch,
            batch_indices,
            loader_name,
            enable_timing=self.timing,
        )

    def execute(
        self,
        net: Union[Tensor, SplitFeatures],
        *,
        clear_cache: bool = True,
        extract_kwargs_dict: Optional[Dict[str, Dict[str, Any]]] = None,
        head_kwargs_dict: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> tensor_dict_type:
        results: Dict[str, Tensor] = {}
        for key, (transform_key, extractor_key, _) in self.pipes.items():
            if key in self.bypassed_pipes:
                continue
            # transform
            transformed = self._transform_cache.get(transform_key)
            if transformed is None:
                transform = self.transforms[transform_key]
                transformed = net if isinstance(net, Tensor) else transform(net)
                self._transform_cache[transform_key] = transformed
            # extract
            if extract_kwargs_dict is None:
                extract_kwargs_dict = {}
            extracted = self._extractor_cache.get(extractor_key)
            if extracted is None:
                extractor = self.extractors[extractor_key]
                extract_kwargs = extract_kwargs_dict.get(extractor_key, {})
                extracted = extractor(transformed, **extract_kwargs)
                extracted_shape = extracted.shape
                if extractor.flatten_ts:
                    if len(extracted_shape) == 3:
                        extracted = extracted.view(extracted_shape[0], -1)
                self._extractor_cache[extractor_key] = extracted
            # execute
            if head_kwargs_dict is None:
                head_kwargs_dict = {}
            head_kwargs = head_kwargs_dict.get(key, {})
            results[key] = self.heads[key](extracted, **head_kwargs)
        # finalize
        if clear_cache:
            self.clear_execute_cache()
        return results

    def clear_execute_cache(self) -> None:
        self._transform_cache = {}
        self._extractor_cache = {}

    def get_split(self, processed: np.ndarray, device: torch.device) -> SplitFeatures:
        with torch.no_grad():
            return self._split_features(to_torch(processed).to(device), None, None)

    def export_context(self) -> context_error_handler:
        class _(context_error_handler):
            def __init__(self, model: ModelBase):
                self.fast_dndf_settings: Dict[DNDF, bool] = {}

                def _inject(node: Module) -> None:
                    for child in node.children():
                        if isinstance(child, DNDF):
                            self.fast_dndf_settings[child] = child._fast
                        elif isinstance(child, Module):
                            _inject(child)

                _inject(model)

            def __enter__(self) -> None:
                for dndf in self.fast_dndf_settings:
                    dndf._fast = False

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                for dndf, fast in self.fast_dndf_settings.items():
                    dndf._fast = fast

        return _(self)

    def extra_repr(self) -> str:
        pipe_str = "\n".join(
            [f"  ({key}): {' -> '.join(pipe[1:])}" for key, pipe in self.pipes.items()]
        )
        return f"(pipes): Pipes(\n{pipe_str}\n)"

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global model_dict

        def before(cls_: Type) -> None:
            cls_.__identifier__ = name

        return register_core(name, model_dict, before_register=before)

    @classmethod
    def register_pipe(
        cls,
        key: str,
        *,
        transform: str = "default",
        extractor: Optional[str] = None,
        reuse_extractor: bool = True,
        head: Optional[str] = None,
        extractor_config: str = "default",
        head_config: str = "default",
        extractor_meta_scope: Optional[str] = None,
        head_meta_scope: Optional[str] = None,
    ) -> Callable[[Type], Type]:
        if head is None:
            head = key
        elif extractor is None:
            extractor = key
        if extractor is None:
            extractor = "identity"

        def _core(cls_: Type) -> Type:
            assert head is not None
            assert extractor is not None
            cfg = PipeConfig(
                transform,
                extractor,
                reuse_extractor,
                head,
                extractor_config,
                head_config,
                extractor_meta_scope,
                head_meta_scope,
            )
            if cls_.registered_pipes is None:
                cls_.registered_pipes = {key: cfg}
            else:
                cls_.registered_pipes[key] = cfg
            return cls_

        return _core


__all__ = ["ModelBase"]



import torch

import numpy as np
import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import register_core

from ..transform import Dimensions
from ...types import tensor_dict_type
from ...configs import configs_dict
from ...configs import Configs
from ...protocol import DataProtocol
from ...misc.toolkit import LoggingMixinWithRank


head_dict: Dict[str, Type["HeadBase"]] = {}


class HeadConfigs(Configs, metaclass=ABCMeta):
    def __init__(
        self,
        in_dim: int,
        tr_data: DataProtocol,
        tr_weights: Optional[np.ndarray],
        dimensions: Dimensions,
        config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(config)
        self.in_dim = in_dim
        self.tr_data = tr_data
        self.tr_weights = tr_weights
        self.dimensions = dimensions

    @property
    def out_dim(self) -> int:
        out_dim = self.config.get("out_dim")
        if self.tr_data.is_clf:
            default_out_dim = self.tr_data.num_classes
        else:
            default_out_dim = self.tr_data.processed.y.shape[1]
        if out_dim is None:
            out_dim = default_out_dim
        return out_dim

    def inject_dimensions(self, config: Dict[str, Any]) -> None:
        config["in_dim"] = self.in_dim
        config["out_dim"] = self.out_dim
        config["dimensions"] = self.dimensions

    @classmethod
    def get(
        cls,
        scope: str,
        name: str,
        *,
        in_dim: Optional[int] = None,
        tr_data: Optional[DataProtocol] = None,
        tr_weights: Optional[np.ndarray] = None,
        dimensions: Optional[Dimensions] = None,
        **kwargs: Any,
    ) -> "HeadConfigs":
        if in_dim is None:
            raise ValueError("`in_dim` must be provided for `HeadConfigs`")
        if tr_data is None:
            raise ValueError("`tr_data` must be provided for `HeadConfigs`")
        if dimensions is None:
            raise ValueError("`dimensions` must be provided for `HeadConfigs`")
        cfg_type = configs_dict[scope][name]
        if not issubclass(cfg_type, HeadConfigs):
            raise ValueError(f"'{name}' under '{scope}' scope is not `HeadConfigs`")
        return cfg_type(in_dim, tr_data, tr_weights, dimensions, kwargs)


class HeadBase(nn.Module, LoggingMixinWithRank, metaclass=ABCMeta):
    def __init__(self, in_dim: int, out_dim: int, **kwargs: Any):
        super().__init__()
        self.in_dim, self.out_dim = in_dim, out_dim
        self.dimensions: Dimensions = kwargs["dimensions"]

    @abstractmethod
    def forward(self, net: torch.Tensor) -> Union[torch.Tensor, tensor_dict_type]:
        pass

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global head_dict
        return register_core(name, head_dict)

    @classmethod
    def make(cls, name: str, config: Dict[str, Any]) -> "HeadBase":
        return head_dict[name](**config)


__all__ = [
    "head_dict",
    "HeadConfigs",
    "HeadBase",
]



from setuptools import setup, find_packages

VERSION = "0.1.12"
DESCRIPTION = "A minimal Automatic Machine Learning (AutoML) solution for tabular datasets based on PyTorch"
with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-learn",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "mlflow",
        "mypy==0.790",
        "black",
        "onnx",
        "onnxruntime",
        "plotly",
        "optuna>=2.3.0",
        "carefree-ml>=0.1.1",
        "carefree-data>=0.2.2",
        "carefree-toolkit>=0.2.5",
        "dill",
        "future",
        "psutil",
        "cython>=0.29.12",
        "numpy>=1.16.2",
        "scipy>=1.2.1",
        "scikit-learn>=0.23.1",
        "matplotlib>=3.0.3",
    ],
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-learn",
    download_url=f"https://github.com/carefree0910/carefree-learn/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python automl machine-learning solution PyTorch",
)



from .api import *
from .data import *
from .models import *
from .trainer import *
from .pipeline import *
from .protocol import *
from .inference import *
from .configs import Configs
from .protocol import PrefetchLoader
from .misc._api import _remove
from .misc._api import _rmtree
from .modules.heads import HeadBase
from .modules.heads import HeadConfigs
from .modules.extractors import ExtractorBase
from .modules.aggregators import AggregatorBase
from .external_ import *

from pkg_resources import get_distribution

pkg = get_distribution("carefree-learn")
__version__ = pkg.version



import os
import json
import torch

import numpy as np

from typing import *
from onnxruntime import InferenceSession
from cftool.misc import shallow_copy_dict
from cftool.misc import lock_manager
from cftool.misc import Saving

from .types import data_type
from .types import np_dict_type
from .protocol import DataProtocol
from .protocol import PrefetchLoader
from .protocol import SamplerProtocol
from .protocol import InferenceProtocol
from .protocol import DataLoaderProtocol
from .misc.toolkit import to_standard
from .misc.toolkit import eval_context
from .misc.toolkit import LoggingMixinWithRank
from .models.base import ModelBase


class PreProcessor(LoggingMixinWithRank):
    data_folder = "data"
    protocols_file = "protocols.json"
    sampler_config_name = "sampler_config"

    def __init__(
        self,
        data: DataProtocol,
        loader_protocol: str,
        sampler_protocol: str,
        sampler_config: Dict[str, Any],
    ):
        self.data = data
        self.data_protocol = data.__identifier__  # type: ignore
        self.loader_protocol = loader_protocol
        self.sampler_protocol = sampler_protocol
        self.sampler_config = sampler_config

    def make_sampler(
        self,
        data: DataProtocol,
        shuffle: bool,
        sample_weights: Optional[np.ndarray] = None,
    ) -> SamplerProtocol:
        config = shallow_copy_dict(self.sampler_config)
        config["shuffle"] = shuffle
        config["sample_weights"] = sample_weights
        return SamplerProtocol.make(self.sampler_protocol, data, **config)

    def make_inference_loader(
        self,
        x: data_type,
        device: Union[str, torch.device],
        batch_size: int = 256,
        *,
        is_onnx: bool,
        contains_labels: bool = False,
    ) -> PrefetchLoader:
        data = self.data.copy_to(x, None, contains_labels=contains_labels)
        loader = DataLoaderProtocol.make(
            self.loader_protocol,
            batch_size,
            self.make_sampler(data, False),
        )
        return PrefetchLoader(loader, device, is_onnx=is_onnx)

    def save(
        self,
        export_folder: str,
        *,
        compress: bool = True,
        save_data: bool = True,
        retain_data: bool = False,
        remove_original: bool = True,
    ) -> "PreProcessor":
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            Saving.prepare_folder(self, export_folder)
            if save_data:
                data_folder = os.path.join(export_folder, self.data_folder)
                self.data.save(
                    data_folder,
                    retain_data=retain_data,
                    compress=False,
                )
            with open(os.path.join(export_folder, self.protocols_file), "w") as f:
                json.dump(
                    {
                        "data": self.data_protocol,
                        "loader": self.loader_protocol,
                        "sampler": self.sampler_protocol,
                    },
                    f,
                )
            Saving.save_dict(
                self.sampler_config,
                self.sampler_config_name,
                export_folder,
            )
            if compress:
                Saving.compress(abs_folder, remove_original=remove_original)
        return self

    @classmethod
    def load(
        cls,
        export_folder: str,
        *,
        data: Optional[DataProtocol] = None,
        compress: bool = True,
    ) -> "PreProcessor":
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(
                export_folder,
                compress,
                remove_extracted=True,
            ):
                with open(os.path.join(export_folder, cls.protocols_file), "r") as f:
                    protocols = json.load(f)
                data_protocol = protocols["data"]
                loader_protocol = protocols["loader"]
                sampler_protocol = protocols["sampler"]
                if data is None:
                    data_folder = os.path.join(export_folder, cls.data_folder)
                    data_base = DataProtocol.get(data_protocol)
                    data = data_base.load(data_folder, compress=False)
                sampler_cfg = Saving.load_dict(cls.sampler_config_name, export_folder)
        return cls(data, loader_protocol, sampler_protocol, sampler_cfg)


class ONNX:
    def __init__(
        self,
        *,
        model: Optional[ModelBase] = None,
        onnx_config: Optional[Dict[str, Any]] = None,
    ):
        if model is None and onnx_config is None:
            raise ValueError("either `model` or `onnx_config` should be provided")

        self.ort_session: Optional[InferenceSession]
        if onnx_config is not None:
            self.model = None
            onnx_path = onnx_config["onnx_path"]
            self.output_names = onnx_config["output_names"]
            self.output_probabilities = onnx_config["output_probabilities"]
            self.ort_session = InferenceSession(onnx_path)
        else:
            assert model is not None
            self.model = model.cpu()
            device, self.model.device = self.model.device, torch.device("cpu")
            self.ort_session = None
            self.input_sample = self.model.input_sample
            with eval_context(self.model):
                outputs = self.model(self.input_sample)
            self.input_names = list(self.input_sample.keys())
            self.output_names = list(outputs.keys())
            self.output_probabilities = model.output_probabilities
            self.model.device = device
            self.model.to(device)

    def to_onnx(
        self,
        onnx_path: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "ONNX":
        if self.model is None:
            raise ValueError("`model` is not provided")
        kwargs = shallow_copy_dict(kwargs)
        kwargs["input_names"] = self.input_names
        kwargs["output_names"] = self.output_names
        kwargs["opset_version"] = 11
        kwargs["export_params"] = True
        kwargs["do_constant_folding"] = True
        if dynamic_axes is None:
            dynamic_axes = {}
        elif isinstance(dynamic_axes, list):
            dynamic_axes = {axis: f"axis.{axis}" for axis in dynamic_axes}
        dynamic_axes[0] = "batch_size"
        dynamic_axes_settings = {}
        for name in self.input_names + self.output_names:
            dynamic_axes_settings[name] = dynamic_axes
        kwargs["dynamic_axes"] = dynamic_axes_settings
        kwargs["verbose"] = verbose
        model = self.model.cpu()
        with eval_context(model):
            torch.onnx.export(
                model,
                self.input_sample,
                onnx_path,
                **shallow_copy_dict(kwargs),
            )
        model.to(model.device)
        return self

    def inference(self, new_inputs: np_dict_type) -> np_dict_type:
        if self.ort_session is None:
            raise ValueError("`onnx_path` is not provided")
        ort_inputs = {
            node.name: to_standard(new_inputs[node.name])
            for node in self.ort_session.get_inputs()
        }
        return dict(zip(self.output_names, self.ort_session.run(None, ort_inputs)))


class Inference(InferenceProtocol, LoggingMixinWithRank):
    def __init__(
        self,
        preprocessor: PreProcessor,
        *,
        model: Optional[ModelBase] = None,
        binary_config: Optional[Dict[str, Any]] = None,
        onnx_config: Optional[Dict[str, Any]] = None,
        use_binary_threshold: bool = True,
        use_tqdm: bool = True,
    ):
        if model is None and onnx_config is None:
            raise ValueError("either `model` or `onnx_config` should be provided")

        self.use_tqdm = use_tqdm
        self.data = preprocessor.data
        self.preprocessor = preprocessor
        self.use_binary_threshold = use_binary_threshold

        # binary case
        self.is_binary = self.data.num_classes == 2
        if binary_config is None:
            binary_config = {}
        self.inject_binary_config(binary_config)

        # onnx
        self.onnx: Optional[ONNX]
        self.model: Optional[ModelBase]

        if onnx_config is not None:
            if model is not None:
                self.log_msg(
                    "`model` and `onnx_config` are both provided, "
                    "`model` will be ignored"
                )
            self.onnx = ONNX(onnx_config=onnx_config)
            self.model = None
        else:
            self.onnx = None
            self.model = model
            if model is None:
                raise ValueError("either `onnx_config` or `model` should be provided")

    def __str__(self) -> str:
        return f"Inference({self.model if self.model is not None else 'ONNX'})"

    __repr__ = __str__

    def inject_binary_config(self, config: Dict[str, Any]) -> None:
        self.binary_metric = config.get("binary_metric")
        self.binary_threshold = config.get("binary_threshold")


__all__ = [
    "PreProcessor",
    "ONNX",
    "Inference",
]



from typing import Any
from typing import Dict

from ..base import HeadConfigs


@HeadConfigs.register("ddr", "default")
class DefaultDDRConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        return {
            "fetch_q": True,
            "fetch_cdf": True,
            "num_layers": 1,
            "num_blocks": 2,
            "latent_dim": 512,
        }


__all__ = ["DefaultDDRConfig"]



from .base import ModelBase


@ModelBase.register("wnd")
@ModelBase.register_pipe("fcnn", transform="embedding")
@ModelBase.register_pipe("linear", transform="one_hot_only")
class WideAndDeep(ModelBase):
    pass


__all__ = ["WideAndDeep"]



from .auxiliary import EMA
from .encoders import *
from .optimizers import *
from .schedulers import *


__all__ = [
    "EMA",
    "Encoder",
    "EncodingResult",
    "optimizer_dict",
    "register_optimizer",
    "scheduler_dict",
    "register_scheduler",
]



from typing import Any
from typing import Dict

from ....configs import Configs


@Configs.register("rnn", "default")
class DefaultRNNConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return {
            "cell": "GRU",
            "num_layers": 1,
            "cell_config": {
                "batch_first": True,
                "hidden_size": 256,
                "bidirectional": False,
            },
        }


__all__ = ["DefaultRNNConfig"]



import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cfdata.tabular import TabularDataset


param_type = Union[torch.Tensor, torch.nn.Parameter]
data_type = Optional[Union[np.ndarray, List[List[float]], str]]
np_dict_type = Dict[str, Union[np.ndarray, Any]]
tensor_dict_type = Dict[str, Union[torch.Tensor, Any]]
tensor_tuple_type = Tuple[torch.Tensor, torch.Tensor]
general_config_type = Optional[Union[str, Dict[str, Any]]]
predictor_type = Optional[Callable[[int, TabularDataset, np.ndarray], np.ndarray]]
evaluator_type = Optional[Callable[[List[np.ndarray], List[np.ndarray]], np.ndarray]]
prefetch_batch_type = Tuple[tensor_dict_type, Optional[torch.Tensor]]
loader_batch_type = Union[tensor_dict_type, prefetch_batch_type]
losses_type = Union[torch.Tensor, tensor_dict_type]


__all__ = [
    "param_type",
    "data_type",
    "np_dict_type",
    "tensor_dict_type",
    "tensor_tuple_type",
    "general_config_type",
    "predictor_type",
    "evaluator_type",
    "prefetch_batch_type",
    "loader_batch_type",
    "losses_type",
]



from typing import Any
from typing import Dict

from ..base import HeadConfigs


@HeadConfigs.register("fcnn", "default")
class DefaultFCNNConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        in_dim = self.in_dim
        if in_dim > 512:
            hidden_units = [1024, 1024]
        elif in_dim > 256:
            if len(self.tr_data) >= 10000:
                hidden_units = [1024, 1024]
            else:
                hidden_units = [2 * in_dim, 2 * in_dim]
        else:
            num_tr_data = len(self.tr_data)
            if num_tr_data >= 100000:
                hidden_units = [768, 768]
            elif num_tr_data >= 10000:
                hidden_units = [512, 512]
            else:
                hidden_dim = max(64 if num_tr_data >= 1000 else 32, 2 * in_dim)
                hidden_units = [hidden_dim, hidden_dim]
        return {
            "hidden_units": hidden_units,
            "mapping_configs": {},
            "final_mapping_config": {},
        }


@HeadConfigs.register("fcnn", "pruned")
class PrunedFCNNConfig(DefaultFCNNConfig):
    def get_default(self) -> Dict[str, Any]:
        config = super().get_default()
        config["mapping_configs"]["pruner_config"] = {}
        return config


__all__ = [
    "DefaultFCNNConfig",
]
